mod broker_task;
mod message;
mod publication;
mod subscription;

use broker_task::{
    AdvertiseRequestSender, BasicAdvertiseRequest, BasicSubscribeRequest, BrokerTask,
    SubscribeRequestSender,
};
pub use message::Message;
pub use publication::{Publication, PublishError};
pub use subscription::{Subscription, SubscriptionError};

use futures::executor::block_on;
use tokio::sync::mpsc;

#[derive(Clone, Debug, PartialEq)]
#[non_exhaustive]
pub enum OrsbError {
    BrokerClosed,
    NoResponse,
}

#[derive(Debug, Clone)]
pub struct Orsb {
    subscribe_tx: SubscribeRequestSender,
    advertise_tx: AdvertiseRequestSender,
}

impl Orsb {
    pub fn start_new() -> Self {
        Self::start_new2().0
    }

    pub(crate) fn start_new2() -> (Self, tokio::task::JoinHandle<()>) {
        let (subscribe_tx, subscribe_rx) = mpsc::channel(8);
        let (advertise_tx, advertise_rx) = mpsc::channel(8);

        let broker = BrokerTask {
            subscribe_rx,
            advertise_rx,
        };
        let join_handle = tokio::spawn(broker.run());

        let orsb = Orsb {
            subscribe_tx,
            advertise_tx,
        };

        (orsb, join_handle)
    }

    pub async fn subscribe<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        let (subscribe_request, response_rx) = BasicSubscribeRequest::<T>::new();
        self.subscribe_tx
            .send(Box::new(subscribe_request))
            .await
            .or(Err(OrsbError::BrokerClosed))?;
        response_rx.await.or(Err(OrsbError::NoResponse))
    }

    pub async fn advertise<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        let (advertise_request, response_rx) = BasicAdvertiseRequest::<T>::new();
        self.advertise_tx
            .send(Box::new(advertise_request))
            .await
            .or(Err(OrsbError::BrokerClosed))?;
        response_rx.await.or(Err(OrsbError::NoResponse))
    }

    pub fn subscribe_blocking<T: Message>(&mut self) -> Result<Subscription<T>, OrsbError> {
        block_on(self.subscribe::<T>())
    }

    pub fn advertise_blocking<T: Message>(&mut self) -> Result<Publication<T>, OrsbError> {
        block_on(self.advertise::<T>())
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[derive(Clone, Debug, PartialEq)]
    struct TestMsg(u8);

    fn init_logger() {
        let _ = env_logger::try_init();
    }

    #[tokio::test]
    async fn test_sync() {
        init_logger();

        let mut orsb = Orsb::start_new();

        tokio::task::spawn_blocking(move || {
            let mut sub = orsb.subscribe_blocking::<TestMsg>().unwrap();
            let mut publ = orsb.advertise_blocking::<TestMsg>().unwrap();

            publ.publish(TestMsg(10)).unwrap();
            publ.publish(TestMsg(20)).unwrap();
            publ.publish(TestMsg(30)).unwrap();

            assert_eq!(sub.recv_blocking(), Ok(TestMsg(10)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(20)));
            assert_eq!(sub.recv_blocking(), Ok(TestMsg(30)));
            assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_async() {
        init_logger();

        let mut orsb = Orsb::start_new();

        let mut sub = orsb.subscribe::<TestMsg>().await.unwrap();
        let mut publ = orsb.advertise::<TestMsg>().await.unwrap();

        publ.publish(TestMsg(10)).unwrap();
        publ.publish(TestMsg(20)).unwrap();
        publ.publish(TestMsg(30)).unwrap();

        assert_eq!(sub.recv().await, Ok(TestMsg(10)));
        assert_eq!(sub.recv().await, Ok(TestMsg(20)));
        assert_eq!(sub.recv().await, Ok(TestMsg(30)));
        assert_eq!(sub.try_recv(), Err(SubscriptionError::Empty));
    }
}
